import 'package:durt/src/gva/queries.dart';
import 'package:durt/src/gva/transaction.dart';
import 'package:graphql/client.dart';

class Gva {
  final GraphQLClient _client;
  Gva._(this._client);

  /// Main constructor for GVA query object
  factory Gva({required String node}) {
    final _httpLink = HttpLink(node);

    final GraphQLClient client = GraphQLClient(
      cache: GraphQLCache(),
      link: _httpLink,
    );

    return Gva._(client);
  }

  /// Return the balance from a given pubkey, or 0 if null.
  Future<double> balance(String pubkey, {bool ud = false}) async {
    double balance = 0;

    final QueryOptions options = QueryOptions(
      document: gql(getBalance),
      variables: <String, dynamic>{
        'pubkey': pubkey,
      },
    );
    final QueryResult result = await _client.query(options);

    if (result.hasException) {
      print(result.exception.toString());
      throw Exception('Unexpected error happened in graphql request');
    } else if (result.data?['balance'] != null) {
      balance = result.data!['balance']['amount'] / 100;
      if (ud) {
        double currentUd = result.data!['currentUd']['amount'] / 100;
        balance = balance / currentUd;
      }
    }

    return balance.toPrecision(2);
  }

  /// Return the username from a given pubkey.
  Future<String> getUsername(String pubkey) async {
    String username = '';

    final QueryOptions options = QueryOptions(
      document: gql(getId),
      variables: <String, dynamic>{
        'pubkey': pubkey,
      },
    );
    final QueryResult result = await _client.query(options);

    if (result.hasException) {
      print(result.exception.toString());
      throw Exception('Unexpected error happened in graphql request');
    } else if (result.data?['idty']?['username'] != null) {
      username = result.data!['idty']['username'];
    }

    return username;
  }

  /// Pay a given pubkey.
  Future<bool> pay(
      {required String recipient,
      required double amount,
      required String dewif,
      required String password,
      int? derivation,
      String comment = '',
      bool useMempool = false}) async {
    Transaction transaction;

    transaction = Transaction(
        recipient: recipient,
        amount: (amount.toPrecision(2) * 100).toInt(),
        dewif: dewif,
        password: password,
        derivation: derivation ?? -1,
        comment: comment,
        useMempool: useMempool,
        client: _client);

    // Execute transaction
    return await transaction.process();
  }
}

extension Ex on double {
  double toPrecision(int n) => double.parse(toStringAsFixed(n));
}
