import 'dart:convert';
import 'package:durt/durt.dart';
import 'package:durt/src/gva/queries.dart';
import 'package:graphql/client.dart';
import 'package:collection/collection.dart';

class Transaction {
  final String recipient;
  final HdWallet wallet;
  final int derivation;
  final String issuerPubkey;
  final int amount;
  final String comment;
  final GraphQLClient client;
  bool useMempool;
  bool isChange = false;

  Transaction._({
    required this.recipient,
    required this.amount,
    required this.wallet,
    required this.derivation,
    required this.issuerPubkey,
    required this.client,
    this.comment = '',
    this.useMempool = false,
  });

  // Main construtor transform dewif + password to hdwallet
  factory Transaction({
    required String recipient,
    required int amount,
    required String dewif,
    required String password,
    required GraphQLClient client,
    required int derivation,
    String comment = '',
    bool useMempool = false,
  }) {
    // TODO: This take a while ... Problem with DEWIF ? Optimisation ?
    HdWallet wallet = HdWallet.fromDewif(dewif, password);

    return Transaction._(
        recipient: recipient,
        amount: amount,
        wallet: wallet,
        derivation: derivation,
        issuerPubkey: wallet.getPubkey(derivation),
        client: client,
        useMempool: useMempool);
  }

  Future<List> _generateTransactionDocument() async {
    final QueryOptions options = QueryOptions(
      document: gql(genTransDocQuery),
      variables: <String, dynamic>{
        'recipient': recipient,
        'amount': amount,
        'issuer': issuerPubkey,
        'comment': comment,
        'useMempool': useMempool
      },
    );
    final QueryResult result = await client.query(options);

    if (result.hasException) {
      throw GraphQLException('GraphQL error: ${result.exception.toString()}');
    } else {
      return result.data!['genTx'];
    }
  }

  bool _checkTransactionDocument(List transDoc) {
    List<String> issuersList = [];
    List<String> recipientsList = [];
    // List outputsList = [];
    List<int> amountsList = [];
    List<String> commentsList = [];

    // Parse received transaction document
    for (String doc in transDoc) {
      List<String> docList = doc.split('\n');

      docList.asMap().forEach((i, line) {
        if (line == 'Issuers:') {
          issuersList.add(docList[i + 1]);
        } else if (line == 'Outputs:') {
          List<String> outputList = docList[i + 1].split(':');
          amountsList.add(int.parse(outputList[0]));
          recipientsList
              .add(outputList[2].split('SIG(')[1].replaceFirst(')', ''));
        } else if (line.startsWith('Comment: ')) {
          commentsList.add(line.split(': ')[1]);
        }
      });
    }
    // Check if it's a change transaction document
    if (recipientsList.every((element) => element == issuerPubkey)) {
      print("The document contain a change transaction");
      isChange = useMempool = true;
      return true;
    } else if (issuersList.every((element) => element != issuerPubkey) ||
        amountsList.sum != amount ||
        recipientsList.every((element) =>
            element != recipient ||
            commentsList.every((element) => element != comment))) {
      // Check validity of transaction document
      print('The generated document is corrupted');
      return false;
    } else {
      isChange = false;
      return true;
    }
  }

  List _signDocument(List documents) {
    List<String> signedDocuments = [];
    for (String doc in documents) {
      var signature = base64Encode(
          wallet.sign(doc, derivation: derivation).buffer.asUint8List());
      signedDocuments.add(doc + signature);
      // print(signedDocuments);
    }
    return signedDocuments;
  }

  Future<bool> _sendSignedDocuments(List signedDocs) async {
    // print(signedDocs);
    for (String doc in signedDocs) {
      final QueryOptions options = QueryOptions(
        document: gql(sendTransactionDocQuery),
        variables: <String, String>{
          'signedDoc': doc,
        },
      );
      final QueryResult result = await client.query(options);

      if (result.hasException) {
        throw GraphQLException('GraphQL error: ${result.exception.toString()}');
      } else if (isChange) {
        await process();
        return true;
      } else {
        print('Transaction complete !');
        return true;
      }
    }
    return false;
  }

  Future<bool> process() async {
    try {
      final List transDocs = await _generateTransactionDocument();
      if (_checkTransactionDocument(transDocs)) {
        final List signedDocs = _signDocument(transDocs);
        final bool result = await _sendSignedDocuments(signedDocs);
        return result;
      } else {
        print('Transaction document is not valid.');
        return false;
      }
    } on GraphQLException catch (e) {
      print('GraphQL error: ' + e.cause.split('message: ')[1].split(',')[0]);
      return false;
    }
  }
}

class GraphQLException implements Exception {
  String cause;
  GraphQLException(this.cause);
}
